use crate::camera::Camera;
use crate::hittable::{Hittable, Sphere};
use crate::ray::Ray;
use crate::vec3::{Color, Point3, Vec3};
use rand::distributions::{Distribution, Uniform};
use crate::material::{Dielectric, Lambertian, Material, Metal};

mod vec3;
mod ray;
mod hittable;
mod camera;
mod material;

fn main() {
    // Image
    const ASPECT_RATIO: f64 = 16.0 / 9.0;
    const IMAGE_WIDTH: i32 = 400;
    const IMAGE_HEIGHT: i32 = (IMAGE_WIDTH as f64 / ASPECT_RATIO) as i32;
    const SAMPLES_PER_PIXEL: i32 = 10;
    const MAX_DEPTH: i32 = 50;

    // Camera
    let cam = Camera::new(ASPECT_RATIO);

    // World
    let mut world:Vec<Box<dyn Hittable>> = Vec::new();

    let center = Box::new(Sphere {
        center: Point3::new(0.0, 0.0, -1.0),
        radius: 0.5,
        material: Material::Lambertian(Lambertian::new(Color::new(0.7, 0.3, 0.3)))
    });
    let center_glass = Box::new(Sphere {
        center: Point3::new(0.0, 0.0, -1.0),
        radius: 0.5,
        material: Material::Dielectric(Dielectric::new(1.5))
    });
    world.push(center_glass);
    let ground = Sphere {
        center: Point3::new(0.0, -100.5, -1.0),
        radius: 100.0,
        material: Material::Lambertian(Lambertian::new(Color::new(0.8, 0.8, 0.0)))
    };
    world.push(Box::new(ground));
    let left = Box::new(Sphere {
        center: Point3::new(-1.0, 0.0, -1.0),
        radius: 0.5,
        material: Material::Metal(Metal::new(Color::new(0.8, 0.8, 0.8), 0.3))
    });
    let left_glass = Box::new(Sphere {
        center: Point3::new(-1.0, 0.0, -1.0),
        radius: 0.5,
        material: Material::Dielectric(Dielectric::new(1.5))
    });
    world.push(left_glass);
    let right = Box::new(Sphere {
        center: Point3::new(1.0, 0.0, -1.0),
        radius: 0.5,
        material: Material::Metal(Metal::new(Color::new(0.8, 0.6, 0.2), 1.0))
    });
    world.push(right);

    println!("P3\n{} {}\n255", IMAGE_WIDTH, IMAGE_HEIGHT);
    let between = Uniform::from(0.0..1.0);
    let mut rng = rand::thread_rng();
    for j in (0..IMAGE_HEIGHT).rev() {
        eprint!("\rScanlines remaining: {} ", j);
        for i in 0..IMAGE_WIDTH {
            let mut color = Color::default();
            for s in 0..SAMPLES_PER_PIXEL {
                let random_number = between.sample(&mut rng);
                let u = (i as f64 + random_number) / (IMAGE_WIDTH - 1) as f64;
                let v = (j as f64 + random_number) / (IMAGE_HEIGHT - 1) as f64;
                let ray = cam.get_ray(u, v);
                color += ray.pixel_color(&world, MAX_DEPTH);
            }
            color.write_color(SAMPLES_PER_PIXEL);
        }
    }
    eprintln!("\nDone");
}
